/*
 * VsParse.c++
 *
 *     Implementation of classes:
 *        VsParseTree
 *
 * The whole enchilada parse tree
 * 
 * Copyright (C) 1996, Silicon Graphics,  Inc.
 */

#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "VsParse.h"
#include "VsFunction.h"

extern int yyparse(); /* The YACC-generated parser */

VsParseTree *VsParseTree::current = NULL;

VsParseTree::VsParseTree()
{
    functions.truncate(0);
    current = NULL;

    scriptName = NULL;

#ifdef __sgi
    errorCB = NULL;
    cbUserData = NULL;
#endif
}

VsParseTree::~VsParseTree()
{
    for (int i = 0; i < functions.getLength(); i++) {
	VsFunctionDef *f = (VsFunctionDef *)functions[i];
	delete f;
    }
    for (i = 0; i < globalVars.getLength(); i+=2) {
	VsVar *var = (VsVar *)globalVars[i+1];

	var->unref();
    }

    if (scriptName != NULL) {
	free((void *)scriptName);
    }
}

int
VsParseTree::parse(const char *buffer, const char *name)
{
    assert(current == NULL);

    current = this;
    curParseString = buffer;
    parsePosition = 0;
    lineNumber = 0;
    numParseErrors = 0;
    
    if (name != NULL) {
        scriptName = strdup(name);
    }

    int result = yyparse();

    current = NULL;
    curParseString = NULL;

    return (getNumErrors() == 0);
}

void
VsParseTree::addGlobalVar(const char *name, VsVar *var)
{
    globalVars.append((void *)name);
    globalVars.append((void *)var);

    var->ref();
}

VsVar *
VsParseTree::getGlobalVar(const char *name)
{
    // Look through globals:
    for (int i = 0; i < globalVars.getLength(); i+= 2) {
	const char *n = (const char *)globalVars[i];
	if (strcmp(n, name) == 0) {
	    VsVar *v = (VsVar *)globalVars[i+1];
	    return v;
	}
    }
    return NULL;
}

VsFunctionDef *
VsParseTree::beginFunction(const char *name)
{
    VsFunctionDef *result = new VsFunctionDef(name, current);

    // ??? Warn if two functions with same name ???
    current->functions.append(result);
    return result;
}

VsFunctionDef *
VsParseTree::findFunction(const char *name)
{
    for (int i = 0; i < functions.getLength(); i++) {
	VsFunctionDef *f = (VsFunctionDef *)functions[i];
	if (strcmp(f->getName(), name) == 0)
	    return f;
    }
    return NULL;
}

void
VsParseTree::yyerror(const char *errmsg)
{
    current->numParseErrors++;

    // Ok, give the user some context about the error:
    // Find beginning and end of current line:
    const char *buf = current->curParseString;
    int lineBegin, lineEnd;
    lineEnd = current->parsePosition;
    lineBegin = lineEnd-1;
    while (lineBegin >= 0 && isprint(buf[lineBegin]))
        --lineBegin;
    ++lineBegin;

    while (isprint(buf[lineEnd])) ++lineEnd;
    --lineEnd;

    int charsInLine = lineEnd-lineBegin + 1;
    int charsInScriptName = 0;
    if (current->scriptName != NULL) {
	charsInScriptName = strlen(current->scriptName);
    }
    char *msg = (char *)malloc(strlen(errmsg)+charsInLine+charsInScriptName+50);

#ifdef __sgi    
    if (current->errorCB == NULL) {
        sprintf(msg, "Script node: %s: line \"", errmsg);
    } else if (current->scriptName != NULL) {
        sprintf(msg, "Script node %s: %s: line %d \"", 
		current->scriptName, 
		errmsg, current->lineNumber+1);	  
    } else {
        sprintf(msg, "Unnamed Script node: %s: line %d \"", errmsg, current->lineNumber+1);
    }
#else 
    sprintf(msg, "Script node: %s: line %d \"", 
		errmsg, current->lineNumber+1);	
#endif
    int msglen = strlen(msg);

    for (int i = 0; i < charsInLine; i++) {
        msg[msglen+i] = buf[lineBegin+i];
    }
    msg[msglen+i] = '"';  // Add close-quote:
    msg[msglen+i+1] = '\n'; // Add new line
    msg[msglen+i+2] = '\0'; // And end-of-string

#ifndef __sgi
     ::MessageBox((HWND)NULL, msg, "VrmlScript syntax error", MB_OK );
#else
    if (current->errorCB != NULL)
        (*(current->errorCB))(current->cbUserData, current->lineNumber+1, msg);
    else
        fprintf(stderr, msg);
#endif

    free(msg);
}
extern "C" {
int yywrap() { return 1; }
};


/* 
	unput

	replace the passed char as the next char in the input stream
*/
void
VsParseTree::unput(int c)
{
    assert(current->parsePosition > 0);
    --current->parsePosition;

    if (c == '\n' || c == '\f') --current->lineNumber;

    // Assertion: always put back what we just got:
    assert(current->curParseString[current->parsePosition] == c);
}

/* 
    input
    get chars from buffer, update wLineNo
*/
int
VsParseTree::input()
{
    char result = current->curParseString[current->parsePosition];
    current->parsePosition++;

    if (result == '\n' || result == '\f') ++current->lineNumber;

    return result;
}
